from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any

from narwhals.translate import get_pandas
from narwhals.translate import get_polars

if TYPE_CHECKING:
    from typing_extensions import Self


class Series:
    def __init__(
        self,
        series: Any,
        *,
        implementation: str | None = None,
    ) -> None:
        from narwhals.pandas_like.series import PandasSeries

        if implementation is not None:
            self._series: Any = series
            self._implementation = implementation
            return
        if (pl := get_polars()) is not None and isinstance(series, pl.Series):
            self._series = series
            self._implementation = "polars"
            return
        if (pd := get_pandas()) is not None and isinstance(series, pd.Series):
            self._series = PandasSeries(series, implementation="pandas")
            self._implementation = "pandas"
            return
        msg = f"Expected pandas or Polars Series, got: {type(series)}"
        raise TypeError(msg)

    def _extract_native(self, arg: Any) -> Any:
        from narwhals.expression import Expr

        if self._implementation != "polars":
            return arg
        if isinstance(arg, Series):
            return arg._series
        if isinstance(arg, Expr):
            import polars as pl

            return arg._call(pl)
        return arg

    def _from_series(self, series: Any) -> Self:
        return self.__class__(series, implementation=self._implementation)

    def __repr__(self) -> str:  # pragma: no cover
        header = " Narwhals Series                                 "
        length = len(header)
        return (
            "┌"
            + "─" * length
            + "┐\n"
            + f"|{header}|\n"
            + "| Use `narwhals.to_native()` to see native output |\n"
            + "└"
            + "─" * length
            + "┘\n"
        )

    def alias(self, name: str) -> Self:
        return self._from_series(self._series.alias(name))

    @property
    def name(self) -> str:
        return self._series.name  # type: ignore[no-any-return]

    @property
    def dtype(self) -> Any:
        return self._series.dtype

    @property
    def shape(self) -> tuple[int]:
        return self._series.shape  # type: ignore[no-any-return]

    def rename(self, name: str) -> Self:
        return self._from_series(self._series.rename(name))

    def cast(
        self,
        dtype: Any,
    ) -> Self:
        return self._from_series(self._series.cast(dtype))

    def item(self) -> Any:
        return self._series.item()

    def is_between(
        self, lower_bound: Any, upper_bound: Any, closed: str = "both"
    ) -> Self:
        return self._from_series(
            self._series.is_between(lower_bound, upper_bound, closed)
        )

    def is_in(self, other: Any) -> Self:
        return self._from_series(self._series.is_in(self._extract_native(other)))

    def is_null(self) -> Self:
        return self._from_series(self._series.is_null())

    def drop_nulls(self) -> Self:
        return self._from_series(self._series.drop_nulls())

    def n_unique(self) -> int:
        return self._series.n_unique()  # type: ignore[no-any-return]

    def unique(self) -> Self:
        return self._from_series(self._series.unique())

    def zip_with(self, mask: Self, other: Self) -> Self:
        return self._from_series(
            self._series.zip_with(self._extract_native(mask), self._extract_native(other))
        )

    def sample(self, n: int, fraction: float, *, with_replacement: bool) -> Self:
        return self._from_series(
            self._series.sample(n, fraction=fraction, with_replacement=with_replacement)
        )

    def to_numpy(self) -> Any:
        return self._series.to_numpy()

    def to_pandas(self) -> Any:
        return self._series.to_pandas()

    def mean(self) -> Any:
        return self._series.mean()

    def std(self) -> Any:
        return self._series.std()

    def __gt__(self, other: Any) -> Series:
        return self._series.__gt__(self._extract_native(other))  # type: ignore[no-any-return]

    def __lt__(self, other: Any) -> Series:
        return self._series.__lt__(self._extract_native(other))  # type: ignore[no-any-return]



# ruff: noqa
from datetime import datetime
from typing import Any

import polars

from narwhals import translate_frame


def q5(
    region_ds_raw: Any,
    nation_ds_raw: Any,
    customer_ds_raw: Any,
    lineitem_ds_raw: Any,
    orders_ds_raw: Any,
    supplier_ds_raw: Any,
) -> Any:
    var_1 = "ASIA"
    var_2 = datetime(1994, 1, 1)
    var_3 = datetime(1995, 1, 1)

    region_ds, pl = translate_frame(region_ds_raw, is_lazy=True)
    nation_ds, _ = translate_frame(nation_ds_raw, is_lazy=True)
    customer_ds, _ = translate_frame(customer_ds_raw, is_lazy=True)
    line_item_ds, _ = translate_frame(lineitem_ds_raw, is_lazy=True)
    orders_ds, _ = translate_frame(orders_ds_raw, is_lazy=True)
    supplier_ds, _ = translate_frame(supplier_ds_raw, is_lazy=True)

    result = (
        region_ds.join(nation_ds, left_on="r_regionkey", right_on="n_regionkey")
        .join(customer_ds, left_on="n_nationkey", right_on="c_nationkey")
        .join(orders_ds, left_on="c_custkey", right_on="o_custkey")
        .join(line_item_ds, left_on="o_orderkey", right_on="l_orderkey")
        .join(
            supplier_ds,
            left_on=["l_suppkey", "n_nationkey"],
            right_on=["s_suppkey", "s_nationkey"],
        )
        .filter(pl.col("r_name") == var_1)
        .filter(pl.col("o_orderdate").is_between(var_2, var_3, closed="left"))
        .with_columns(
            (pl.col("l_extendedprice") * (1 - pl.col("l_discount"))).alias("revenue")
        )
        .group_by("n_name")
        .agg([pl.sum("revenue")])
        .sort(by="revenue", descending=True)
    )

    return result.collect().to_native()


region_ds = polars.scan_parquet("../tpch-data/region.parquet")
nation_ds = polars.scan_parquet("../tpch-data/nation.parquet")
customer_ds = polars.scan_parquet("../tpch-data/customer.parquet")
lineitem_ds = polars.scan_parquet("../tpch-data/lineitem.parquet")
orders_ds = polars.scan_parquet("../tpch-data/orders.parquet")
supplier_ds = polars.scan_parquet("../tpch-data/supplier.parquet")
print(
    q5(
        region_ds.collect().to_pandas(),
        nation_ds.collect().to_pandas(),
        customer_ds.collect().to_pandas(),
        lineitem_ds.collect().to_pandas(),
        orders_ds.collect().to_pandas(),
        supplier_ds.collect().to_pandas(),
    )
)
print(
    q5(
        region_ds,
        nation_ds,
        customer_ds,
        lineitem_ds,
        orders_ds,
        supplier_ds,
    )
)



from __future__ import annotations

from typing import Any

import polars as pl


def compare_dicts(result: dict[str, Any], expected: dict[str, Any]) -> None:
    if isinstance(result, pl.LazyFrame):
        result = result.collect()
    for key in expected:
        for lhs, rhs in zip(result[key], expected[key]):
            if isinstance(lhs, float):
                assert abs(lhs - rhs) < 1e-6
            else:
                assert lhs == rhs



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any

if TYPE_CHECKING:
    from narwhals.pandas_like.dataframe import PandasDataFrame
    from narwhals.pandas_like.series import PandasSeries


def translate_frame(
    df: Any,
    implementation: str,
) -> PandasDataFrame:
    from narwhals.pandas_like.dataframe import PandasDataFrame

    return PandasDataFrame(
        df,
        implementation=implementation,
    )


def translate_series(
    series: Any,
    implementation: str,
) -> PandasSeries:
    from narwhals.pandas_like.series import PandasSeries

    return PandasSeries(series, implementation=implementation)



# mypy: ignore
# ruff: noqa
import re
import subprocess
import sys

how = sys.argv[1]

with open("pyproject.toml", encoding="utf-8") as f:
    content = f.read()
old_version = re.search(r'version = "(.*)"', content).group(1)
version = old_version.split(".")
if how == "patch":
    version = ".".join(version[:-1] + [str(int(version[-1]) + 1)])
elif how == "minor":
    version = ".".join(version[:-2] + [str(int(version[-2]) + 1), "0"])
elif how == "major":
    version = ".".join([str(int(version[0]) + 1), "0", "0"])
content = content.replace(f'version = "{old_version}"', f'version = "{version}"')
with open("pyproject.toml", "w", encoding="utf-8") as f:
    f.write(content)

with open("narwhals/__init__.py", encoding="utf-8") as f:
    content = f.read()
content = content.replace(
    f'__version__ = "{old_version}"',
    f'__version__ = "{version}"',
)
with open("narwhals/__init__.py", "w", encoding="utf-8") as f:
    f.write(content)

subprocess.run(["git", "commit", "-a", "-m", f"Bump version to {version}"])
subprocess.run(["git", "tag", "-a", version, "-m", version])
subprocess.run(["git", "push", "--follow-tags"])



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any

from narwhals.pandas_like.utils import isinstance_or_issubclass

if TYPE_CHECKING:
    from typing_extensions import Self


class DType:
    def __repr__(self) -> str:
        return self.__class__.__qualname__

    @classmethod
    def is_numeric(cls: type[Self]) -> bool:
        return issubclass(cls, NumericType)

    def __eq__(self, other: DType | type[DType]) -> bool:  # type: ignore[override]
        return isinstance_or_issubclass(other, type(self))


class NumericType(DType): ...


class TemporalType(DType): ...


class Int64(NumericType): ...


class Int32(NumericType): ...


class Int16(NumericType): ...


class Int8(NumericType): ...


class UInt64(NumericType): ...


class UInt32(NumericType): ...


class UInt16(NumericType): ...


class UInt8(NumericType): ...


class Float64(NumericType): ...


class Float32(NumericType): ...


class String(DType): ...


class Boolean(DType): ...


class Object(DType):  # todo: do we really want this one?
    ...


class Datetime(TemporalType): ...


class Date(TemporalType): ...


def translate_dtype(plx: Any, dtype: DType) -> Any:
    if dtype == Float64:
        return plx.Float64
    if dtype == Float32:
        return plx.Float32
    if dtype == Int64:
        return plx.Int64
    if dtype == Int32:
        return plx.Int32
    if dtype == Int16:
        return plx.Int16
    if dtype == UInt8:
        return plx.UInt8
    if dtype == UInt64:
        return plx.UInt64
    if dtype == UInt32:
        return plx.UInt32
    if dtype == UInt16:
        return plx.UInt16
    if dtype == UInt8:
        return plx.UInt8
    if dtype == String:
        return plx.String
    if dtype == Boolean:
        return plx.Boolean
    if dtype == Datetime:
        return plx.Datetime
    if dtype == Date:
        return plx.Date
    msg = f"Unknown dtype: {dtype}"
    raise TypeError(msg)


def to_narwhals_dtype(dtype: Any, implementation: str) -> DType:
    if implementation != "polars":
        return dtype  # type: ignore[no-any-return]
    import polars as pl

    if dtype == pl.Float64:
        return Float64()
    if dtype == pl.Float32:
        return Float32()
    if dtype == pl.Int64:
        return Int64()
    if dtype == pl.Int32:
        return Int32()
    if dtype == pl.Int16:
        return Int16()
    if dtype == pl.UInt8:
        return UInt8()
    if dtype == pl.UInt64:
        return UInt64()
    if dtype == pl.UInt32:
        return UInt32()
    if dtype == pl.UInt16:
        return UInt16()
    if dtype == pl.UInt8:
        return UInt8()
    if dtype == pl.String:
        return String()
    if dtype == pl.Boolean:
        return Boolean()
    if dtype == pl.Datetime:
        return Datetime()
    if dtype == pl.Date:
        return Date()
    msg = f"Unexpected dtype, got: {type(dtype)}"
    raise TypeError(msg)



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any
from typing import Sequence

from pandas.api.types import is_extension_array_dtype

from narwhals.pandas_like.utils import item
from narwhals.pandas_like.utils import reverse_translate_dtype
from narwhals.pandas_like.utils import translate_dtype
from narwhals.pandas_like.utils import validate_column_comparand

if TYPE_CHECKING:
    from typing_extensions import Self

    from narwhals.dtypes import DType


class PandasSeries:
    def __init__(
        self,
        series: Any,
        *,
        implementation: str,
    ) -> None:
        """Parameters
        ----------
        df
            DataFrame this column originates from.
        """

        self._name = str(series.name) if series.name is not None else ""
        self._series = series
        self._implementation = implementation

    def _from_series(self, series: Any) -> Self:
        return self.__class__(
            series.rename(series.name, copy=False),
            implementation=self._implementation,
        )

    def __len__(self) -> int:
        return self.shape[0]

    @property
    def name(self) -> str:
        return self._name

    @property
    def shape(self) -> tuple[int]:
        return self._series.shape  # type: ignore[no-any-return]

    def rename(self, name: str) -> PandasSeries:
        ser = self._series
        return self._from_series(ser.rename(name, copy=False))

    @property
    def dtype(self) -> DType:
        return translate_dtype(self._series.dtype)

    def cast(
        self,
        dtype: Any,
    ) -> Self:
        ser = self._series
        dtype = reverse_translate_dtype(dtype)
        return self._from_series(ser.astype(dtype))

    def filter(self, mask: Self) -> Self:
        ser = self._series
        return self._from_series(
            ser.loc[validate_column_comparand(self._series.index, mask)]
        )

    def item(self) -> Any:
        return item(self._series)

    def is_between(
        self, lower_bound: Any, upper_bound: Any, closed: str = "both"
    ) -> PandasSeries:
        ser = self._series
        return self._from_series(ser.between(lower_bound, upper_bound, inclusive=closed))

    def is_in(self, other: Any) -> PandasSeries:
        import pandas as pd

        ser = self._series
        res = ser.isin(other).convert_dtypes()
        res[ser.isna()] = pd.NA
        return self._from_series(res)

    # Binary comparisons

    def __eq__(self, other: object) -> PandasSeries:  # type: ignore[override]
        ser = self._series
        other = validate_column_comparand(self._series.index, other)
        return self._from_series((ser.__eq__(other)).rename(ser.name, copy=False))

    def __ne__(self, other: object) -> PandasSeries:  # type: ignore[override]
        ser = self._series
        other = validate_column_comparand(self._series.index, other)
        return self._from_series((ser.__ne__(other)).rename(ser.name, copy=False))

    def __ge__(self, other: Any) -> PandasSeries:
        ser = self._series
        other = validate_column_comparand(self._series.index, other)
        return self._from_series((ser.__ge__(other)).rename(ser.name, copy=False))

    def __gt__(self, other: Any) -> PandasSeries:
        ser = self._series
        other = validate_column_comparand(self._series.index, other)
        return self._from_series((ser.__gt__(other)).rename(ser.name, copy=False))

    def __le__(self, other: Any) -> PandasSeries:
        ser = self._series
        other = validate_column_comparand(self._series.index, other)
        return self._from_series((ser.__le__(other)).rename(ser.name, copy=False))

    def __lt__(self, other: Any) -> PandasSeries:
        ser = self._series
        other = validate_column_comparand(self._series.index, other)
        return self._from_series((ser.__lt__(other)).rename(ser.name, copy=False))

    def __and__(self, other: Any) -> PandasSeries:
        ser = self._series
        other = validate_column_comparand(self._series.index, other)
        return self._from_series((ser.__and__(other)).rename(ser.name, copy=False))

    def __rand__(self, other: Any) -> PandasSeries:
        ser = self._series
        other = validate_column_comparand(self._series.index, other)
        return self._from_series((ser.__rand__(other)).rename(ser.name, copy=False))

    def __or__(self, other: Any) -> PandasSeries:
        ser = self._series
        other = validate_column_comparand(self._series.index, other)
        return self._from_series((ser.__or__(other)).rename(ser.name, copy=False))

    def __ror__(self, other: Any) -> PandasSeries:
        ser = self._series
        other = validate_column_comparand(self._series.index, other)
        return self._from_series((ser.__ror__(other)).rename(ser.name, copy=False))

    def __add__(self, other: Any) -> PandasSeries:
        ser = self._series
        other = validate_column_comparand(self._series.index, other)
        return self._from_series((ser.__add__(other)).rename(ser.name, copy=False))

    def __radd__(self, other: Any) -> PandasSeries:
        ser = self._series
        other = validate_column_comparand(self._series.index, other)
        return self._from_series((ser.__radd__(other)).rename(ser.name, copy=False))

    def __sub__(self, other: Any) -> PandasSeries:
        ser = self._series
        other = validate_column_comparand(self._series.index, other)
        return self._from_series((ser.__sub__(other)).rename(ser.name, copy=False))

    def __rsub__(self, other: Any) -> PandasSeries:
        ser = self._series
        other = validate_column_comparand(self._series.index, other)
        return self._from_series((ser.__rsub__(other)).rename(ser.name, copy=False))

    def __mul__(self, other: Any) -> PandasSeries:
        ser = self._series
        other = validate_column_comparand(self._series.index, other)
        return self._from_series((ser.__mul__(other)).rename(ser.name, copy=False))

    def __rmul__(self, other: Any) -> PandasSeries:
        ser = self._series
        other = validate_column_comparand(self._series.index, other)
        return self._from_series((ser.__rmul__(other)).rename(ser.name, copy=False))

    def __truediv__(self, other: Any) -> PandasSeries:
        ser = self._series
        other = validate_column_comparand(self._series.index, other)
        return self._from_series((ser.__truediv__(other)).rename(ser.name, copy=False))

    def __rtruediv__(self, other: Any) -> PandasSeries:
        ser = self._series
        other = validate_column_comparand(self._series.index, other)
        return self._from_series((ser.__rtruediv__(other)).rename(ser.name, copy=False))

    def __floordiv__(self, other: Any) -> PandasSeries:
        ser = self._series
        other = validate_column_comparand(self._series.index, other)
        return self._from_series((ser.__floordiv__(other)).rename(ser.name, copy=False))

    def __rfloordiv__(self, other: Any) -> PandasSeries:
        ser = self._series
        other = validate_column_comparand(self._series.index, other)
        return self._from_series((ser.__rfloordiv__(other)).rename(ser.name, copy=False))

    def __pow__(self, other: Any) -> PandasSeries:
        ser = self._series
        other = validate_column_comparand(self._series.index, other)
        return self._from_series((ser.__pow__(other)).rename(ser.name, copy=False))

    def __rpow__(self, other: Any) -> PandasSeries:  # pragma: no cover
        ser = self._series
        other = validate_column_comparand(self._series.index, other)
        return self._from_series((ser.__rpow__(other)).rename(ser.name, copy=False))

    def __mod__(self, other: Any) -> PandasSeries:
        ser = self._series
        other = validate_column_comparand(self._series.index, other)
        return self._from_series((ser.__mod__(other)).rename(ser.name, copy=False))

    def __rmod__(self, other: Any) -> PandasSeries:  # pragma: no cover
        ser = self._series
        other = validate_column_comparand(self._series.index, other)
        return self._from_series((ser.__rmod__(other)).rename(ser.name, copy=False))

    # Unary

    def __invert__(self: PandasSeries) -> PandasSeries:
        ser = self._series
        return self._from_series(~ser)

    # Reductions

    def any(self) -> Any:
        ser = self._series
        return ser.any()

    def all(self) -> Any:
        ser = self._series
        return ser.all()

    def min(self) -> Any:
        ser = self._series
        return ser.min()

    def max(self) -> Any:
        ser = self._series
        return ser.max()

    def sum(self) -> Any:
        ser = self._series
        return ser.sum()

    def prod(self) -> Any:
        ser = self._series
        return ser.prod()

    def median(self) -> Any:
        ser = self._series
        return ser.median()

    def mean(self) -> Any:
        ser = self._series
        return ser.mean()

    def std(
        self,
        *,
        correction: float = 1.0,
    ) -> Any:
        ser = self._series
        return ser.std(ddof=correction)

    def var(
        self,
        *,
        correction: float = 1.0,
    ) -> Any:
        ser = self._series
        return ser.var(ddof=correction)

    def len(self) -> Any:
        return len(self._series)

    # Transformations

    def is_null(self) -> PandasSeries:
        ser = self._series
        return self._from_series(ser.isna())

    def drop_nulls(self) -> PandasSeries:
        ser = self._series
        return self._from_series(ser.dropna())

    def n_unique(self) -> int:
        ser = self._series
        return ser.nunique()  # type: ignore[no-any-return]

    def zip_with(self, mask: PandasSeries, other: PandasSeries) -> PandasSeries:
        mask = validate_column_comparand(self._series.index, mask)
        other = validate_column_comparand(self._series.index, other)
        ser = self._series
        return self._from_series(ser.where(mask, other))

    def sample(self, n: int, fraction: float, *, with_replacement: bool) -> PandasSeries:
        ser = self._series
        return self._from_series(
            ser.sample(n=n, frac=fraction, with_replacement=with_replacement)
        )

    def unique(self) -> PandasSeries:
        if self._implementation != "pandas":
            raise NotImplementedError
        import pandas as pd

        return self._from_series(
            pd.Series(
                self._series.unique(), dtype=self._series.dtype, name=self._series.name
            )
        )

    def is_nan(self) -> PandasSeries:
        ser = self._series
        if is_extension_array_dtype(ser.dtype):
            return self._from_series((ser != ser).fillna(False))  # noqa: PLR0124
        return self._from_series(ser.isna())

    def sort(
        self,
        *,
        descending: bool | Sequence[bool] = True,
    ) -> PandasSeries:
        ser = self._series
        return self._from_series(
            ser.sort_values(ascending=not descending).rename(self.name)
        )

    def alias(self, name: str) -> Self:
        ser = self._series
        return self._from_series(ser.rename(name, copy=False))

    def to_numpy(self) -> Any:
        return self._series.to_numpy()

    def to_pandas(self) -> Any:
        if self._implementation == "pandas":
            return self._series
        elif self._implementation == "cudf":
            return self._series.to_pandas()
        elif self._implementation == "modin":
            return self._series._to_pandas()
        msg = f"Unknown implementation: {self._implementation}"
        raise TypeError(msg)



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import Iterable

from narwhals.dtypes import translate_dtype
from narwhals.utils import flatten

if TYPE_CHECKING:
    from narwhals.typing import IntoExpr


def extract_native(expr: Expr, other: Any) -> Any:
    if isinstance(other, Expr):
        return other._call(expr)
    return other


class Expr:
    def __init__(self, call: Callable[[Any], Any]) -> None:
        # callable from namespace to expr
        self._call = call

    # --- convert ---
    def alias(self, name: str) -> Expr:
        return self.__class__(lambda plx: self._call(plx).alias(name))

    def cast(
        self,
        dtype: Any,
    ) -> Expr:
        return self.__class__(
            lambda plx: self._call(plx).cast(translate_dtype(plx, dtype)),
        )

    # --- binary ---
    def __eq__(self, other: object) -> Expr:  # type: ignore[override]
        return self.__class__(
            lambda plx: self._call(plx).__eq__(extract_native(plx, other))
        )

    def __and__(self, other: Any) -> Expr:
        return self.__class__(
            lambda plx: self._call(plx).__and__(extract_native(plx, other))
        )

    def __or__(self, other: Any) -> Expr:
        return self.__class__(
            lambda plx: self._call(plx).__or__(extract_native(plx, other))
        )

    def __add__(self, other: Any) -> Expr:
        return self.__class__(
            lambda plx: self._call(plx).__add__(extract_native(plx, other))
        )

    def __radd__(self, other: Any) -> Expr:
        return self.__class__(
            lambda plx: self._call(plx).__radd__(extract_native(plx, other))
        )

    def __sub__(self, other: Any) -> Expr:
        return self.__class__(
            lambda plx: self._call(plx).__sub__(extract_native(plx, other))
        )

    def __rsub__(self, other: Any) -> Expr:
        return self.__class__(
            lambda plx: self._call(plx).__rsub__(extract_native(plx, other))
        )

    def __truediv__(self, other: Any) -> Expr:
        return self.__class__(
            lambda plx: self._call(plx).__truediv__(extract_native(plx, other))
        )

    def __rtruediv__(self, other: Any) -> Expr:
        return self.__class__(
            lambda plx: self._call(plx).__rtruediv__(extract_native(plx, other))
        )

    def __mul__(self, other: Any) -> Expr:
        return self.__class__(
            lambda plx: self._call(plx).__mul__(extract_native(plx, other))
        )

    def __rmul__(self, other: Any) -> Expr:
        return self.__class__(
            lambda plx: self._call(plx).__rmul__(extract_native(plx, other))
        )

    def __le__(self, other: Any) -> Expr:
        return self.__class__(
            lambda plx: self._call(plx).__le__(extract_native(plx, other))
        )

    def __lt__(self, other: Any) -> Expr:
        return self.__class__(
            lambda plx: self._call(plx).__lt__(extract_native(plx, other))
        )

    def __gt__(self, other: Any) -> Expr:
        return self.__class__(
            lambda plx: self._call(plx).__gt__(extract_native(plx, other))
        )

    def __ge__(self, other: Any) -> Expr:
        return self.__class__(
            lambda plx: self._call(plx).__ge__(extract_native(plx, other))
        )

    # --- unary ---
    def mean(self) -> Expr:
        return self.__class__(lambda plx: self._call(plx).mean())

    def sum(self) -> Expr:
        return self.__class__(lambda plx: self._call(plx).sum())

    def min(self) -> Expr:
        return self.__class__(lambda plx: self._call(plx).min())

    def max(self) -> Expr:
        return self.__class__(lambda plx: self._call(plx).max())

    def n_unique(self) -> Expr:
        return self.__class__(lambda plx: self._call(plx).n_unique())

    def unique(self) -> Expr:
        return self.__class__(lambda plx: self._call(plx).unique())

    # --- transform ---
    def is_between(
        self, lower_bound: Any, upper_bound: Any, closed: str = "both"
    ) -> Expr:
        return self.__class__(
            lambda plx: self._call(plx).is_between(lower_bound, upper_bound, closed)
        )

    def is_in(self, other: Any) -> Expr:
        return self.__class__(lambda plx: self._call(plx).is_in(other))

    def is_null(self) -> Expr:
        return self.__class__(lambda plx: self._call(plx).is_null())

    # --- partial reduction ---
    def drop_nulls(self) -> Expr:
        return self.__class__(lambda plx: self._call(plx).drop_nulls())

    def sample(
        self,
        n: int | None = None,
        fraction: float | None = None,
        *,
        with_replacement: bool = False,
    ) -> Expr:
        return self.__class__(
            lambda plx: self._call(plx).sample(
                n, fraction=fraction, with_replacement=with_replacement
            )
        )


def col(*names: str | Iterable[str]) -> Expr:
    return Expr(lambda plx: plx.col(*names))


def all() -> Expr:
    return Expr(lambda plx: plx.all())


def len() -> Expr:
    return Expr(lambda plx: plx.len())


def sum(*columns: str) -> Expr:
    return Expr(lambda plx: plx.sum(*columns))


def mean(*columns: str) -> Expr:
    return Expr(lambda plx: plx.mean(*columns))


def min(*columns: str) -> Expr:
    return Expr(lambda plx: plx.min(*columns))


def max(*columns: str) -> Expr:
    return Expr(lambda plx: plx.max(*columns))


def sum_horizontal(*exprs: IntoExpr | Iterable[IntoExpr]) -> Expr:
    return Expr(
        lambda plx: plx.sum_horizontal([extract_native(plx, v) for v in flatten(exprs)])
    )


__all__ = [
    "Expr",
]



from __future__ import annotations

import os
import warnings
from typing import Any

import numpy as np
import pandas as pd
import polars as pl
import pytest

import narwhals as nw
from tests.utils import compare_dicts

df_pandas = pd.DataFrame({"a": [1, 3, 2], "b": [4, 4, 6], "z": [7.0, 8, 9]})
df_polars = pl.DataFrame({"a": [1, 3, 2], "b": [4, 4, 6], "z": [7.0, 8, 9]})
df_lazy = pl.LazyFrame({"a": [1, 3, 2], "b": [4, 4, 6], "z": [7.0, 8, 9]})

if os.environ.get("CI", None):
    import modin.pandas as mpd

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning)
        df_mpd = mpd.DataFrame(
            pd.DataFrame({"a": [1, 3, 2], "b": [4, 4, 6], "z": [7.0, 8, 9]})
        )
else:
    df_mpd = df_pandas.copy()


@pytest.mark.parametrize(
    "df_raw",
    [df_pandas, df_polars, df_lazy],
)
def test_sort(df_raw: Any) -> None:
    df = nw.LazyFrame(df_raw)
    result = df.sort("a", "b")
    result_native = nw.to_native(result)
    expected = {
        "a": [1, 2, 3],
        "b": [4, 6, 4],
        "z": [7.0, 9.0, 8.0],
    }
    compare_dicts(result_native, expected)


@pytest.mark.parametrize(
    "df_raw",
    [df_pandas, df_lazy],
)
def test_filter(df_raw: Any) -> None:
    df = nw.LazyFrame(df_raw)
    result = df.filter(nw.col("a") > 1)
    result_native = nw.to_native(result)
    expected = {"a": [3, 2], "b": [4, 6], "z": [8.0, 9.0]}
    compare_dicts(result_native, expected)


@pytest.mark.parametrize(
    "df_raw",
    [df_pandas, df_polars],
)
def test_filter_series(df_raw: Any) -> None:
    df = nw.DataFrame(df_raw).with_columns(mask=nw.col("a") > 1)
    result = df.filter(df["mask"]).drop("mask")
    result_native = nw.to_native(result)
    expected = {"a": [3, 2], "b": [4, 6], "z": [8.0, 9.0]}
    compare_dicts(result_native, expected)


@pytest.mark.parametrize(
    "df_raw",
    [df_pandas, df_lazy],
)
def test_add(df_raw: Any) -> None:
    df = nw.LazyFrame(df_raw)
    result = df.with_columns(
        c=nw.col("a") + nw.col("b"),
        d=nw.col("a") - nw.col("a").mean(),
    )
    result_native = nw.to_native(result)
    expected = {
        "a": [1, 3, 2],
        "b": [4, 4, 6],
        "z": [7.0, 8.0, 9.0],
        "c": [5, 7, 8],
        "d": [-1.0, 1.0, 0.0],
    }
    compare_dicts(result_native, expected)


@pytest.mark.parametrize(
    "df_raw",
    [df_pandas, df_lazy],
)
def test_double(df_raw: Any) -> None:
    df = nw.LazyFrame(df_raw)
    result = df.with_columns(nw.all() * 2)
    result_native = nw.to_native(result)
    expected = {"a": [2, 6, 4], "b": [8, 8, 12], "z": [14.0, 16.0, 18.0]}
    compare_dicts(result_native, expected)


@pytest.mark.parametrize(
    "df_raw",
    [df_pandas, df_lazy],
)
def test_select(df_raw: Any) -> None:
    df = nw.LazyFrame(df_raw)
    result = df.select("a")
    result_native = nw.to_native(result)
    expected = {"a": [1, 3, 2]}
    compare_dicts(result_native, expected)


@pytest.mark.parametrize("df_raw", [df_pandas, df_lazy])
def test_sumh(df_raw: Any) -> None:
    df = nw.LazyFrame(df_raw)
    result = df.with_columns(horizonal_sum=nw.sum_horizontal(nw.col("a"), nw.col("b")))
    result_native = nw.to_native(result)
    expected = {
        "a": [1, 3, 2],
        "b": [4, 4, 6],
        "z": [7.0, 8.0, 9.0],
        "horizonal_sum": [5, 7, 8],
    }
    compare_dicts(result_native, expected)


@pytest.mark.parametrize("df_raw", [df_pandas, df_lazy])
def test_sumh_literal(df_raw: Any) -> None:
    df = nw.LazyFrame(df_raw)
    result = df.with_columns(horizonal_sum=nw.sum_horizontal("a", nw.col("b")))
    result_native = nw.to_native(result)
    expected = {
        "a": [1, 3, 2],
        "b": [4, 4, 6],
        "z": [7.0, 8.0, 9.0],
        "horizonal_sum": [5, 7, 8],
    }
    compare_dicts(result_native, expected)


@pytest.mark.parametrize("df_raw", [df_pandas, df_lazy])
def test_sum_all(df_raw: Any) -> None:
    df = nw.LazyFrame(df_raw)
    result = df.select(nw.all().sum())
    result_native = nw.to_native(result)
    expected = {"a": [6], "b": [14], "z": [24.0]}
    compare_dicts(result_native, expected)


@pytest.mark.parametrize("df_raw", [df_pandas, df_lazy])
def test_double_selected(df_raw: Any) -> None:
    df = nw.LazyFrame(df_raw)
    result = df.select(nw.col("a", "b") * 2)
    result_native = nw.to_native(result)
    expected = {"a": [2, 6, 4], "b": [8, 8, 12]}
    compare_dicts(result_native, expected)


@pytest.mark.parametrize("df_raw", [df_pandas, df_lazy])
def test_rename(df_raw: Any) -> None:
    df = nw.LazyFrame(df_raw)
    result = df.rename({"a": "x", "b": "y"})
    result_native = nw.to_native(result)
    expected = {"x": [1, 3, 2], "y": [4, 4, 6], "z": [7.0, 8, 9]}
    compare_dicts(result_native, expected)


@pytest.mark.parametrize("df_raw", [df_pandas, df_lazy])
def test_join(df_raw: Any) -> None:
    df = nw.LazyFrame(df_raw)
    df_right = df.rename({"z": "z_right"})
    result = df.join(df_right, left_on=["a", "b"], right_on=["a", "b"], how="inner")
    result_native = nw.to_native(result)
    expected = {"a": [1, 3, 2], "b": [4, 4, 6], "z": [7.0, 8, 9], "z_right": [7.0, 8, 9]}
    compare_dicts(result_native, expected)


@pytest.mark.parametrize("df_raw", [df_pandas, df_lazy])
def test_schema(df_raw: Any) -> None:
    df = nw.LazyFrame(df_raw)
    result = df.schema
    expected = {"a": nw.dtypes.Int64, "b": nw.dtypes.Int64, "z": nw.dtypes.Float64}
    assert result == expected


@pytest.mark.parametrize("df_raw", [df_pandas, df_lazy])
def test_columns(df_raw: Any) -> None:
    df = nw.LazyFrame(df_raw)
    result = df.columns
    expected = ["a", "b", "z"]
    assert result == expected


@pytest.mark.parametrize("df_raw", [df_polars, df_pandas, df_mpd, df_lazy])
def test_lazy_instantiation(df_raw: Any) -> None:
    result = nw.LazyFrame(df_raw)
    result_native = nw.to_native(result)
    expected = {"a": [1, 3, 2], "b": [4, 4, 6], "z": [7.0, 8, 9]}
    compare_dicts(result_native, expected)


@pytest.mark.parametrize("df_raw", [df_lazy])
def test_lazy_instantiation_error(df_raw: Any) -> None:
    with pytest.raises(
        TypeError, match="Can't instantiate DataFrame from Polars LazyFrame."
    ):
        _ = nw.DataFrame(df_raw).shape


@pytest.mark.parametrize("df_raw", [df_polars, df_pandas, df_mpd])
def test_eager_instantiation(df_raw: Any) -> None:
    result = nw.DataFrame(df_raw)
    result_native = nw.to_native(result)
    expected = {"a": [1, 3, 2], "b": [4, 4, 6], "z": [7.0, 8, 9]}
    compare_dicts(result_native, expected)


def test_accepted_dataframes() -> None:
    array = np.array([[0, 4.0], [2, 5]])
    with pytest.raises(
        TypeError,
        match="Expected pandas-like dataframe, Polars dataframe, or Polars lazyframe, got: <class 'numpy.ndarray'>",
    ):
        nw.DataFrame(array)
    with pytest.raises(
        TypeError,
        match="Expected pandas-like dataframe, Polars dataframe, or Polars lazyframe, got: <class 'numpy.ndarray'>",
    ):
        nw.LazyFrame(array)


@pytest.mark.parametrize("df_raw", [df_polars, df_pandas, df_mpd])
@pytest.mark.filterwarnings("ignore:.*Passing a BlockManager.*:DeprecationWarning")
def test_convert_pandas(df_raw: Any) -> None:
    result = nw.DataFrame(df_raw).to_pandas()
    expected = pd.DataFrame({"a": [1, 3, 2], "b": [4, 4, 6], "z": [7.0, 8, 9]})
    pd.testing.assert_frame_equal(result, expected)


@pytest.mark.parametrize("df_raw", [df_polars, df_pandas, df_mpd])
@pytest.mark.filterwarnings(
    r"ignore:np\.find_common_type is deprecated\.:DeprecationWarning"
)
def test_convert_numpy(df_raw: Any) -> None:
    result = nw.DataFrame(df_raw).to_numpy()
    expected = np.array([[1, 3, 2], [4, 4, 6], [7.0, 8, 9]]).T
    np.testing.assert_array_equal(result, expected)


@pytest.mark.parametrize("df_raw", [df_polars, df_pandas, df_mpd])
def test_shape(df_raw: Any) -> None:
    result = nw.DataFrame(df_raw).shape
    expected = (3, 3)
    assert result == expected


@pytest.mark.parametrize("df_raw", [df_polars, df_pandas, df_mpd, df_lazy])
def test_expr_binary(df_raw: Any) -> None:
    result = nw.LazyFrame(df_raw).with_columns(
        a=(1 + 3 * nw.col("a")) * (1 / nw.col("a")),
        b=nw.col("z") / (2 - nw.col("b")),
        c=nw.col("a") + nw.col("b") / 2,
        d=nw.col("a") - nw.col("b"),
        e=((nw.col("a") > nw.col("b")) & (nw.col("a") >= nw.col("z"))).cast(nw.Int64),
        f=(
            (nw.col("a") < nw.col("b"))
            | (nw.col("a") <= nw.col("z"))
            | (nw.col("a") == 1)
        ).cast(nw.Int64),
    )
    result_native = nw.to_native(result)
    expected = {
        "a": [4, 3.333333, 3.5],
        "b": [-3.5, -4.0, -2.25],
        "z": [7.0, 8.0, 9.0],
        "c": [3, 5, 5],
        "d": [-3, -1, -4],
        "e": [0, 0, 0],
        "f": [1, 1, 1],
    }
    compare_dicts(result_native, expected)


@pytest.mark.parametrize("df_raw", [df_polars, df_pandas, df_lazy])
def test_expr_unary(df_raw: Any) -> None:
    result = (
        nw.LazyFrame(df_raw)
        .with_columns(
            a_mean=nw.col("a").mean(),
            a_sum=nw.col("a").sum(),
            b_nunique=nw.col("b").n_unique(),
            z_min=nw.col("z").min(),
            z_max=nw.col("z").max(),
        )
        .select(nw.col("a_mean", "a_sum", "b_nunique", "z_min", "z_max").unique())
    )
    result_native = nw.to_native(result)
    expected = {"a_mean": [2], "a_sum": [6], "b_nunique": [2], "z_min": [7], "z_max": [9]}
    compare_dicts(result_native, expected)


@pytest.mark.parametrize("df_raw", [df_polars, df_pandas, df_mpd, df_lazy])
def test_expr_transform(df_raw: Any) -> None:
    result = nw.LazyFrame(df_raw).with_columns(
        a=nw.col("a").is_between(-1, 1), b=nw.col("b").is_in([4, 5])
    )
    result_native = nw.to_native(result)
    expected = {"a": [True, False, False], "b": [True, True, False], "z": [7, 8, 9]}
    compare_dicts(result_native, expected)


@pytest.mark.parametrize("df_raw", [df_polars, df_pandas, df_lazy])
def test_expr_min_max(df_raw: Any) -> None:
    df = nw.LazyFrame(df_raw)
    result_min = nw.to_native(df.select(nw.min("a", "b", "z")))
    result_max = nw.to_native(df.select(nw.max("a", "b", "z")))
    expected_min = {"a": [1], "b": [4], "z": [7]}
    expected_max = {"a": [3], "b": [6], "z": [9]}
    compare_dicts(result_min, expected_min)
    compare_dicts(result_max, expected_max)



from __future__ import annotations

from typing import Any
from typing import Iterable


def remove_prefix(text: str, prefix: str) -> str:
    if text.startswith(prefix):
        return text[len(prefix) :]
    return text


def flatten(args: Any) -> list[Any]:
    if not args:
        return []
    if len(args) == 1 and _is_iterable(args[0]):
        return args[0]  # type: ignore[no-any-return]
    return args  # type: ignore[no-any-return]


def _is_iterable(arg: Any | Iterable[Any]) -> bool:
    from narwhals.series import Series

    return isinstance(arg, Iterable) and not isinstance(arg, (str, bytes, Series))


def flatten_str(*args: str | Iterable[str]) -> list[str]:
    out: list[str] = []
    for arg in args:
        if isinstance(arg, str):
            out.append(arg)
        else:
            for item in arg:
                if not isinstance(item, str):
                    msg = f"Expected str, got {type(item)}"
                    raise TypeError(msg)
                out.append(item)
    return out


def flatten_bool(*args: bool | Iterable[bool]) -> list[bool]:
    out: list[bool] = []
    for arg in args:
        if isinstance(arg, bool):
            out.append(arg)
        else:
            for item in arg:
                if not isinstance(item, bool):
                    msg = f"Expected str, got {type(item)}"
                    raise TypeError(msg)
                out.append(item)
    return out



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any
from typing import Iterable
from typing import Literal
from typing import Sequence

from narwhals.dtypes import to_narwhals_dtype
from narwhals.pandas_like.dataframe import PandasDataFrame
from narwhals.translate import get_cudf
from narwhals.translate import get_modin
from narwhals.translate import get_pandas
from narwhals.translate import get_polars

if TYPE_CHECKING:
    from typing_extensions import Self

    from narwhals.dtypes import DType
    from narwhals.group_by import GroupBy
    from narwhals.series import Series
    from narwhals.typing import IntoExpr


class BaseFrame:
    _dataframe: Any
    _implementation: str

    def _from_dataframe(self, df: Any) -> Self:
        # construct, preserving properties
        return self.__class__(  # type: ignore[call-arg]
            df,
            implementation=self._implementation,
        )

    def _flatten_and_extract(self, *args: Any, **kwargs: Any) -> Any:
        from narwhals.utils import flatten

        args = [self._extract_native(v) for v in flatten(args)]  # type: ignore[assignment]
        kwargs = {k: self._extract_native(v) for k, v in kwargs.items()}
        return args, kwargs

    def _extract_native(self, arg: Any) -> Any:
        from narwhals.expression import Expr
        from narwhals.pandas_like.namespace import PandasNamespace
        from narwhals.series import Series

        if isinstance(arg, BaseFrame):
            return arg._dataframe
        if isinstance(arg, Series):
            return arg._series
        if isinstance(arg, Expr):
            if self._implementation == "polars":
                import polars as pl

                return arg._call(pl)
            plx = PandasNamespace(implementation=self._implementation)
            return arg._call(plx)
        return arg

    def __repr__(self) -> str:  # pragma: no cover
        header = " Narwhals DataFrame                              "
        length = len(header)
        return (
            "┌"
            + "─" * length
            + "┐\n"
            + f"|{header}|\n"
            + "| Use `narwhals.to_native()` to see native output |\n"
            + "└"
            + "─" * length
            + "┘\n"
        )

    @property
    def schema(self) -> dict[str, DType]:
        return {
            k: to_narwhals_dtype(v, self._implementation)
            for k, v in self._dataframe.schema.items()
        }

    @property
    def columns(self) -> list[str]:
        return self._dataframe.columns  # type: ignore[no-any-return]

    def with_columns(
        self, *exprs: IntoExpr | Iterable[IntoExpr], **named_exprs: IntoExpr
    ) -> Self:
        exprs, named_exprs = self._flatten_and_extract(*exprs, **named_exprs)
        return self._from_dataframe(
            self._dataframe.with_columns(*exprs, **named_exprs),
        )

    def select(
        self,
        *exprs: IntoExpr | Iterable[IntoExpr],
        **named_exprs: IntoExpr,
    ) -> Self:
        exprs, named_exprs = self._flatten_and_extract(*exprs, **named_exprs)
        return self._from_dataframe(
            self._dataframe.select(*exprs, **named_exprs),
        )

    def rename(self, mapping: dict[str, str]) -> Self:
        return self._from_dataframe(self._dataframe.rename(mapping))

    def drop(self, *columns: str | Iterable[str]) -> Self:
        return self._from_dataframe(self._dataframe.drop(*columns))

    def filter(self, *predicates: IntoExpr | Iterable[IntoExpr]) -> Self:
        predicates, _ = self._flatten_and_extract(*predicates)
        return self._from_dataframe(
            self._dataframe.filter(*predicates),
        )

    def group_by(self, *keys: str | Iterable[str]) -> GroupBy:
        from narwhals.group_by import GroupBy

        # todo: groupby and lazygroupby
        return GroupBy(self, *keys)  # type: ignore[arg-type]

    def sort(
        self,
        by: str | Iterable[str],
        *more_by: str,
        descending: bool | Sequence[bool] = False,
    ) -> Self:
        return self._from_dataframe(
            self._dataframe.sort(by, *more_by, descending=descending)
        )

    def join(
        self,
        other: Self,
        *,
        how: Literal["inner"] = "inner",
        left_on: str | list[str],
        right_on: str | list[str],
    ) -> Self:
        return self._from_dataframe(
            self._dataframe.join(
                self._extract_native(other),
                how=how,
                left_on=left_on,
                right_on=right_on,
            )
        )


class DataFrame(BaseFrame):
    def __init__(
        self,
        df: Any,
        *,
        implementation: str | None = None,
    ) -> None:
        if implementation is not None:
            self._dataframe: Any = df
            self._implementation = implementation
            return
        if (pl := get_polars()) is not None and isinstance(df, pl.DataFrame):
            self._dataframe = df
            self._implementation = "polars"
        elif (pl := get_polars()) is not None and isinstance(df, pl.LazyFrame):
            raise TypeError(
                "Can't instantiate DataFrame from Polars LazyFrame. Call `collect()` first, or use `narwhals.LazyFrame` if you don't specifically require eager execution."
            )
        elif (pd := get_pandas()) is not None and isinstance(df, pd.DataFrame):
            self._dataframe = PandasDataFrame(df, implementation="pandas")
            self._implementation = "pandas"
        elif (mpd := get_modin()) is not None and isinstance(
            df, mpd.DataFrame
        ):  # pragma: no cover
            self._dataframe = PandasDataFrame(df, implementation="modin")
            self._implementation = "modin"
        elif (cudf := get_cudf()) is not None and isinstance(
            df, cudf.DataFrame
        ):  # pragma: no cover
            self._dataframe = PandasDataFrame(df, implementation="cudf")
            self._implementation = "cudf"
        elif hasattr(df, "__narwhals_dataframe__"):  # pragma: no cover
            self._dataframe = df.__narwhals_dataframe__()
            self._implementation = "custom"
        else:
            msg = f"Expected pandas-like dataframe, Polars dataframe, or Polars lazyframe, got: {type(df)}"
            raise TypeError(msg)

    def to_pandas(self) -> Any:
        return self._dataframe.to_pandas()

    def to_numpy(self) -> Any:
        return self._dataframe.to_numpy()

    @property
    def shape(self) -> tuple[int, int]:
        return self._dataframe.shape  # type: ignore[no-any-return]

    def __getitem__(self, col_name: str) -> Series:
        from narwhals.series import Series

        return Series(self._dataframe[col_name], implementation=self._implementation)

    def to_dict(self, *, as_series: bool = True) -> dict[str, Any]:
        return self._dataframe.to_dict(as_series=as_series)  # type: ignore[no-any-return]


class LazyFrame(BaseFrame):
    def __init__(
        self,
        df: Any,
        *,
        implementation: str | None = None,
    ) -> None:
        if implementation is not None:
            self._dataframe: Any = df
            self._implementation = implementation
            return
        if (pl := get_polars()) is not None and isinstance(
            df, (pl.DataFrame, pl.LazyFrame)
        ):
            self._dataframe = df.lazy()
            self._implementation = "polars"
        elif (pd := get_pandas()) is not None and isinstance(df, pd.DataFrame):
            self._dataframe = PandasDataFrame(df, implementation="pandas")
            self._implementation = "pandas"
        elif (mpd := get_modin()) is not None and isinstance(
            df, mpd.DataFrame
        ):  # pragma: no cover
            self._dataframe = PandasDataFrame(df, implementation="modin")
            self._implementation = "modin"
        elif (cudf := get_cudf()) is not None and isinstance(
            df, cudf.DataFrame
        ):  # pragma: no cover
            self._dataframe = PandasDataFrame(df, implementation="cudf")
            self._implementation = "cudf"
        else:
            msg = f"Expected pandas-like dataframe, Polars dataframe, or Polars lazyframe, got: {type(df)}"
            raise TypeError(msg)

    def collect(self) -> DataFrame:
        return DataFrame(
            self._dataframe.collect(),
            implementation=self._implementation,
        )



from typing import TYPE_CHECKING
from typing import TypeAlias
from typing import TypeVar

if TYPE_CHECKING:
    from narwhals.expression import Expr
    from narwhals.series import Series

    IntoExpr: TypeAlias = Expr | str | int | float | Series

    NativeDataFrame = TypeVar("NativeDataFrame")
    NativeSeries = TypeVar("NativeSeries")

T = TypeVar("T")



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any

from narwhals.dependencies import get_cudf
from narwhals.dependencies import get_modin
from narwhals.dependencies import get_pandas
from narwhals.dependencies import get_polars

if TYPE_CHECKING:
    from narwhals.dataframe import BaseFrame
    from narwhals.series import Series


def to_native(obj: BaseFrame | Series) -> Any:
    from narwhals.dataframe import BaseFrame
    from narwhals.series import Series

    if isinstance(obj, BaseFrame):
        return (
            obj._dataframe
            if obj._implementation == "polars"
            else obj._dataframe._dataframe
        )
    if isinstance(obj, Series):
        return obj._series if obj._implementation == "polars" else obj._series._series

    msg = f"Expected Narwhals object, got {type(obj)}."
    raise TypeError(msg)


__all__ = [
    "get_pandas",
    "get_polars",
    "get_modin",
    "get_cudf",
    "to_native",
]



from narwhals.containers import get_implementation
from narwhals.containers import is_dataframe
from narwhals.containers import is_pandas
from narwhals.containers import is_polars
from narwhals.containers import is_series
from narwhals.dataframe import DataFrame
from narwhals.dataframe import LazyFrame
from narwhals.dtypes import *  # noqa: F403
from narwhals.expression import all
from narwhals.expression import col
from narwhals.expression import len
from narwhals.expression import max
from narwhals.expression import mean
from narwhals.expression import min
from narwhals.expression import sum
from narwhals.expression import sum_horizontal
from narwhals.series import Series
from narwhals.translate import to_native

__version__ = "0.6.5"

__all__ = [
    "is_dataframe",
    "is_series",
    "is_polars",
    "is_pandas",
    "get_implementation",
    "to_native",
    "all",
    "col",
    "len",
    "min",
    "max",
    "mean",
    "sum",
    "sum_horizontal",
    "DataFrame",
    "LazyFrame",
    "Series",
]



from __future__ import annotations

import collections
import os
import warnings
from copy import copy
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import Iterable

from packaging.version import parse

from narwhals.pandas_like.utils import dataframe_from_dict
from narwhals.pandas_like.utils import evaluate_simple_aggregation
from narwhals.pandas_like.utils import horizontal_concat
from narwhals.pandas_like.utils import is_simple_aggregation
from narwhals.pandas_like.utils import item
from narwhals.pandas_like.utils import parse_into_exprs
from narwhals.utils import remove_prefix

if TYPE_CHECKING:
    from narwhals.pandas_like.dataframe import PandasDataFrame
    from narwhals.pandas_like.expr import PandasExpr
    from narwhals.pandas_like.typing import IntoPandasExpr


class PandasGroupBy:
    def __init__(self, df: PandasDataFrame, keys: list[str]) -> None:
        self._df = df
        self._keys = list(keys)

    def agg(
        self,
        *aggs: IntoPandasExpr | Iterable[IntoPandasExpr],
        **named_aggs: IntoPandasExpr,
    ) -> PandasDataFrame:
        df = self._df._dataframe
        exprs = parse_into_exprs(
            self._df._implementation,
            *aggs,
            **named_aggs,
        )
        grouped = df.groupby(
            list(self._keys),
            sort=False,
            as_index=False,
        )
        implementation: str = self._df._implementation
        output_names: list[str] = copy(self._keys)
        for expr in exprs:
            if expr._output_names is None:
                msg = (
                    "Anonymous expressions are not supported in group_by.agg.\n"
                    "Instead of `pl.all()`, try using a named expression, such as "
                    "`pl.col('a', 'b')`\n"
                )
                raise ValueError(msg)
            output_names.extend(expr._output_names)

        if implementation in ("pandas", "modin") and not os.environ.get(
            "NARWHALS_FORCE_GENERIC"
        ):
            return agg_pandas(
                grouped,
                exprs,
                self._keys,
                output_names,
                self._from_dataframe,
            )
        return agg_generic(
            grouped,
            exprs,
            self._keys,
            output_names,
            implementation,
            self._from_dataframe,
        )

    def _from_dataframe(self, df: PandasDataFrame) -> PandasDataFrame:
        from narwhals.pandas_like.dataframe import PandasDataFrame

        return PandasDataFrame(
            df,
            implementation=self._df._implementation,
        )


def agg_pandas(
    grouped: Any,
    exprs: list[PandasExpr],
    keys: list[str],
    output_names: list[str],
    from_dataframe: Callable[[Any], PandasDataFrame],
) -> PandasDataFrame:
    """
    This should be the fastpath, but cuDF is too far behind to use it.

    - https://github.com/rapidsai/cudf/issues/15118
    - https://github.com/rapidsai/cudf/issues/15084
    """
    import pandas as pd

    simple_aggs = []
    complex_aggs = []
    for expr in exprs:
        if is_simple_aggregation(expr):
            simple_aggs.append(expr)
        else:
            complex_aggs.append(expr)
    simple_aggregations = {}
    for expr in simple_aggs:
        if expr._depth == 0:
            # e.g. agg(pl.len())
            assert expr._output_names is not None
            for output_name in expr._output_names:
                simple_aggregations[output_name] = pd.NamedAgg(
                    column=keys[0], aggfunc=expr._function_name.replace("len", "size")
                )
            continue

        assert expr._root_names is not None
        assert expr._output_names is not None
        for root_name, output_name in zip(expr._root_names, expr._output_names):
            name = remove_prefix(expr._function_name, "col->")
            simple_aggregations[output_name] = pd.NamedAgg(column=root_name, aggfunc=name)
    result_simple = grouped.agg(**simple_aggregations) if simple_aggregations else None

    def func(df: Any) -> Any:
        out_group = []
        out_names = []
        for expr in complex_aggs:
            results_keys = expr._call(from_dataframe(df))
            for result_keys in results_keys:
                out_group.append(item(result_keys._series))
                out_names.append(result_keys.name)
        return pd.Series(out_group, index=out_names)

    if complex_aggs:
        warnings.warn(
            "Found complex group-by expression, which can't be expressed efficiently with the "
            "pandas API. If you can, please rewrite your query such that group-by aggregations "
            "are simple (e.g. mean, std, min, max, ...).",
            UserWarning,
            stacklevel=2,
        )
        if parse(pd.__version__) < parse("2.2.0"):
            result_complex = grouped.apply(func)
        else:
            result_complex = grouped.apply(func, include_groups=False)

    if result_simple is not None and not complex_aggs:
        result = result_simple
    elif result_simple is not None and complex_aggs:
        result = pd.concat(
            [result_simple, result_complex.drop(columns=keys)],
            axis=1,
            copy=False,
        )
    elif complex_aggs:
        result = result_complex
    else:
        raise AssertionError("At least one aggregation should have been passed")
    return from_dataframe(result.loc[:, output_names])


def agg_generic(  # noqa: PLR0913
    grouped: Any,
    exprs: list[PandasExpr],
    group_by_keys: list[str],
    output_names: list[str],
    implementation: str,
    from_dataframe: Callable[[Any], PandasDataFrame],
) -> PandasDataFrame:
    dfs: list[Any] = []
    to_remove: list[int] = []
    for i, expr in enumerate(exprs):
        if is_simple_aggregation(expr):
            dfs.append(evaluate_simple_aggregation(expr, grouped, group_by_keys))
            to_remove.append(i)
    exprs = [expr for i, expr in enumerate(exprs) if i not in to_remove]

    out: dict[str, list[Any]] = collections.defaultdict(list)
    for keys, df_keys in grouped:
        for key, name in zip(keys, group_by_keys):
            out[name].append(key)
        for expr in exprs:
            results_keys = expr._call(from_dataframe(df_keys))
            for result_keys in results_keys:
                out[result_keys.name].append(result_keys.item())

    results_keys = dataframe_from_dict(out, implementation=implementation)
    results_df = horizontal_concat(
        [results_keys, *dfs], implementation=implementation
    ).loc[:, output_names]
    return from_dataframe(results_df)



from typing import Any

from narwhals.translate import get_pandas
from narwhals.translate import get_polars


def is_dataframe(obj: Any) -> bool:
    if (pl := get_polars()) is not None and isinstance(obj, (pl.DataFrame, pl.LazyFrame)):
        return True
    if (pd := get_pandas()) is not None and isinstance(obj, pd.DataFrame):
        return True
    return False


def is_series(obj: Any) -> bool:
    if (pl := get_polars()) is not None and isinstance(obj, (pl.Series)):
        return True
    if (pd := get_pandas()) is not None and isinstance(obj, pd.Series):
        return True
    raise NotImplementedError


def get_implementation(obj: Any) -> str:
    if (pl := get_polars()) is not None and isinstance(
        obj, (pl.DataFrame, pl.LazyFrame, pl.Expr, pl.Series)
    ):
        return "polars"
    if (pd := get_pandas()) is not None and isinstance(obj, (pd.DataFrame, pd.Series)):
        return "pandas"
    msg = f"Unknown implementation: {obj}"
    raise TypeError(msg)


def is_pandas(obj: Any) -> bool:
    return get_implementation(obj) == "pandas"


def is_polars(obj: Any) -> bool:
    return get_implementation(obj) == "polars"


def is_cudf(obj: Any) -> bool:
    return get_implementation(obj) == "cudf"


def is_modin(obj: Any) -> bool:
    return get_implementation(obj) == "modin"



# ruff: noqa
from datetime import datetime
from typing import Any

import polars

from narwhals import translate_frame

Q_NUM = 4


def q4(
    lineitem_ds_raw: Any,
    orders_ds_raw: Any,
) -> Any:
    var_1 = datetime(1993, 7, 1)
    var_2 = datetime(1993, 10, 1)

    line_item_ds, pl = translate_frame(lineitem_ds_raw, is_lazy=True)
    orders_ds, _ = translate_frame(orders_ds_raw, is_lazy=True)

    result = (
        line_item_ds.join(orders_ds, left_on="l_orderkey", right_on="o_orderkey")
        .filter(pl.col("o_orderdate").is_between(var_1, var_2, closed="left"))
        .filter(pl.col("l_commitdate") < pl.col("l_receiptdate"))
        .unique(subset=["o_orderpriority", "l_orderkey"])
        .group_by("o_orderpriority")
        .agg(pl.len().alias("order_count"))
        .sort(by="o_orderpriority")
        .with_columns(
            pl.col("order_count")
            # .cast(pl.datatypes.Int64)
        )
    )

    return result.collect().to_native()


lineitem_ds = polars.scan_parquet("../tpch-data/lineitem.parquet")
orders_ds = polars.scan_parquet("../tpch-data/orders.parquet")
print(
    q4(
        lineitem_ds.collect().to_pandas(),
        orders_ds.collect().to_pandas(),
    )
)
print(
    q4(
        lineitem_ds,
        orders_ds,
    )
)



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import TypeAlias

if TYPE_CHECKING:
    from narwhals.pandas_like.expr import PandasExpr
    from narwhals.pandas_like.series import PandasSeries

    IntoPandasExpr: TypeAlias = PandasExpr | str | int | float | PandasSeries



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Iterable

if TYPE_CHECKING:
    from narwhals.dataframe import DataFrame
    from narwhals.dataframe import LazyFrame
    from narwhals.typing import IntoExpr

# todo: make groupby and lazygroupby


class GroupBy:
    def __init__(self, df: DataFrame | LazyFrame, *keys: str | Iterable[str]) -> None:
        self._df = df
        self._keys = keys

    def agg(
        self, *aggs: IntoExpr | Iterable[IntoExpr], **named_aggs: IntoExpr
    ) -> DataFrame | LazyFrame:
        aggs, named_aggs = self._df._flatten_and_extract(*aggs, **named_aggs)
        return self._df.__class__(
            self._df._dataframe.group_by(*self._keys).agg(*aggs, **named_aggs),
            implementation=self._df._implementation,
        )
